-module(echo).

-export([start/0,
         stop/0]).

-export([reactor/1]).

-define(TWS_TCP_PORT, 8080).
-define(TWS_RECV_IDLE_TIMEOUT, 3000).

-record(reactor_state, {
        client_sock % TWS socket
        }).


start() ->
    tws:start_link({local, ?MODULE}, ?TWS_TCP_PORT, {?MODULE, reactor}).

% stop the GAP proxy
stop() ->
    tws:stop(?MODULE).


%%====================================================================
%% INTERNAL FUNCTIONS
%%====================================================================
reactor(ClientSocket) ->
    State = #reactor_state{client_sock = ClientSocket},
    ok = tws_loop(State).

tws_loop(#reactor_state{client_sock = ClientSocket} = State) ->
   case gen_tcp:recv(ClientSocket, 0, ?TWS_RECV_IDLE_TIMEOUT) of
       {ok, Chunk} ->
           % echo back what we received
           gen_tcp:send(ClientSocket, Chunk),
           tws_loop(State);
       {error, timeout} ->
           io:format("+++ Timeout (Pid: ~p)~n", [self()]),
           gen_tcp:close(ClientSocket),
           ok;
       
       {error, closed} ->
           ok
    end.

