-module(tws).

-behaviour(gen_server).

-export([start_link/3,
         stop/1]).

-export([init/1, handle_call/3, handle_cast/2, handle_info/2,
         terminate/2, code_change/3]).

-export([accept_loop/1]).

-define(TWS_TCP_BACKLOG,   128).
-define(TWS_TCP_OPTIONS, [binary, 
                               {packet,    0},
                               {reuseaddr, true},
                               {active,    false},
                               {backlog,   ?TWS_TCP_BACKLOG}]).


-record(server_state, {
        port,             % listening port
        loop,             % callback function
        lsocket = null}). % local socket


-define(SERVER, ?MODULE).


%--------------------------------------------------------------------
start_link(ServerName, Port, Loop) ->
    State = #server_state{port = Port, loop = Loop},
    gen_server:start_link(ServerName, ?MODULE, State, []).

%%--------------------------------------------------------------------
stop(ServerName) ->
       gen_server:cast(ServerName, stop).


%%====================================================================
%% gen_server CALLBACKS
%%====================================================================
init(State = #server_state{port=Port}) ->
    case gen_tcp:listen(Port, ?TWS_TCP_OPTIONS) of
        {ok, LSocket} ->
            NewState = State#server_state{lsocket = LSocket},
            {ok, accept(NewState)};
        {error, Reason} ->
            {stop, Reason}
    end.

%%--------------------------------------------------------------------
handle_call(_Request, _From, State) ->
    Reply = ok,
    {reply, Reply, State}.

%%--------------------------------------------------------------------
handle_cast(stop, State) ->
    {stop, normal, State};

%%--------------------------------------------------------------------
handle_cast({accepted, _Pid}, State=#server_state{}) ->
    {noreply, accept(State)};

%%--------------------------------------------------------------------
handle_cast(_Msg, State) ->
    {noreply, State}.


%%--------------------------------------------------------------------
handle_info(_Info, State) ->
    {noreply, State}.

%%--------------------------------------------------------------------
terminate(_Reason, _State) ->
    ok.

%%--------------------------------------------------------------------
code_change(_OldVsn, State, _Extra) ->
    {ok, State}.



%%====================================================================
%% INTERNAL FUNCTIONS
%%====================================================================
accept(State = #server_state{lsocket = LSocket, loop = Loop}) ->
    proc_lib:spawn_link(?MODULE, accept_loop, [{self(), LSocket, Loop}]),
    State.

accept_loop({Server, LSocket, {M, F}}) ->
    {ok, Socket} = gen_tcp:accept(LSocket),
    gen_server:cast(Server, {accepted, self()}),
    M:F(Socket).
